import torch, scipy

from componentScale import trainHandler

def eval_logit_change(interv_logits_NIT:torch.tensor, orig_logits_NIT:torch.tensor, answ_toks_N2:torch.tensor, probs:bool=False):
    if probs: ## turn logits into probabilities
        orig_logits_NIT, interv_logits_NIT = torch.softmax(orig_logits_NIT, dim=-1), torch.softmax(interv_logits_NIT, dim=-1)
    orig_logit_diff, _ = trainHandler.get_logit_diff(orig_logits_NIT, answ_toks_N2=answ_toks_N2)
    interv_logit_diff, (interv_logit_tok0, interv_logit_tok1) = trainHandler.get_logit_diff(interv_logits_NIT, answ_toks_N2=answ_toks_N2)
    logit_change = (orig_logit_diff - interv_logit_diff).mean()
    return logit_change, (interv_logit_tok0, interv_logit_tok1)

def eval_token_ranks(logits_NIT:torch.tensor, answ_toks_N2:torch.tensor=[]):
    logits_NIT, answ_toks_N2 = logits_NIT.cpu(), answ_toks_N2.cpu()
    idx_ranks_NT = scipy.stats.rankdata(-logits_NIT[:,-1,:], axis=-1, method='min')-1
    n_indcs = torch.arange(0, answ_toks_N2.shape[0])
    interv_rank_tok0, interv_rank_tok1 = (idx_ranks_NT[n_indcs,answ_toks_N2[...,0]]), (idx_ranks_NT[n_indcs,answ_toks_N2[...,1]])
    return (interv_rank_tok0, interv_rank_tok1)
    
def eval_intervention(model, hooks=None, toks_NI:list=[], answ_toks_N2:list=[]):
    torch.set_grad_enabled(False)
    logit_changes, logit_tok0, logit_tok1, tok0_rank, tok1_rank = [], [], [], [], []
    for (toks_I, answ_toks_2) in zip(toks_NI, answ_toks_N2):
        toks_nI, answ_toks_n2 = toks_I.unsqueeze(0), answ_toks_2.unsqueeze(0)
        orig_logits_NIT, _ = model.run_with_cache(toks_nI)
        interv_logits_NIT = model.run_with_hooks(toks_nI, fwd_hooks=hooks, return_type="logits", reset_hooks_end=True)
        logit_change, (interv_logit_tok0, interv_logit_tok1) = eval_logit_change(interv_logits_NIT, orig_logits_NIT, answ_toks_n2)
        (interv_rank_tok0, interv_rank_tok1) = eval_token_ranks(interv_logits_NIT, answ_toks_n2)
        (orig_rank_tok0, orig_rank_tok1) = eval_token_ranks(orig_logits_NIT, answ_toks_n2)
        print(f"logit diff: {round(logit_change.item(), 3)}, tok0: {orig_rank_tok0.item()}->{interv_rank_tok0.item()}, tok1: {orig_rank_tok1.item()}->{interv_rank_tok1.item()}, {model.to_str_tokens(answ_toks_n2.view(-1))}")
        
        logit_changes.append(logit_change), logit_tok0.append(interv_logit_tok0), logit_tok1.append(interv_logit_tok1)
        tok0_rank.append(interv_rank_tok0), tok1_rank.append(interv_rank_tok1)
    
    return logit_changes, (logit_tok0, logit_tok1), (tok0_rank, tok1_rank)

def eval_beta_range(sh, beta_range, eval_fn):
    beta_tok_logits, beta_logit_changes, beta_tok_ranks = [], [], []
    for beta in beta_range:
        sh.beta = beta
        logit_change, tok_logits, tok_ranks = eval_fn(hooks = sh.hooks)
        beta_tok_logits.append(tok_logits)
        beta_logit_changes.append(logit_change)
        beta_tok_ranks.append(tok_ranks)

    beta_tok_logits = torch.tensor(beta_tok_logits).squeeze(-1)
    return beta_range, beta_tok_logits, beta_logit_changes, beta_tok_ranks


def binary_rank_comparison(tok_ranks_pos:list, tok_ranks_neg:list):

    pos_ranks_toks = torch.stack((torch.tensor(tok_ranks_pos[0]), torch.tensor(tok_ranks_pos[1])))
    neg_ranks_toks = torch.stack((torch.tensor(tok_ranks_neg[0]), torch.tensor(tok_ranks_neg[1])))
    
    pos_ranks, pos_tok = torch.min(pos_ranks_toks, dim=0)
    neg_ranks, neg_tok = torch.min(neg_ranks_toks, dim=0)

    top_tok_changed = pos_tok != neg_tok
    top_tok_flipped = top_tok_changed.int().sum() / len(top_tok_changed.int())
    print(f"binary_rank_comparison: {round(top_tok_flipped.item(),3)}")
    return top_tok_flipped
