import seaborn as sns
import numpy as np
import matplotlib as mpl 
import matplotlib.pyplot as plt 
from matplotlib.ticker import FuncFormatter, MaxNLocator, FormatStrFormatter
mpl.rcParams['axes.spines.right'] = False
mpl.rcParams['axes.spines.top'] = False

from componentScale import dataHandler

def plot_attributions(c_type, attrType, SCALARS, model, toks_nI, answ_toks_n2, f_name:str="", show_every:int=-1, figsize=(1.9,2.5), fontsize=11):
    #(2.3,2.75)

    if isinstance(SCALARS[next(iter(SCALARS))], list):
        SCALARS = {k: torch.stack(v).squeeze() for k, v in SCALARS.items()}

    if c_type in ["k","q","v","z"]:
        vals = SCALARS[c_type].detach().numpy()
        yticklabels = [f"L{l}" for l in range(0, vals.shape[0])]
        xticklabels = model.to_str_tokens(toks_nI)
        
        if len(vals.shape)==3: ## heads
            if vals.shape[1] == 1: ## not multiple token positions
                vals = vals.squeeze(1)
                xticklabels = [f"H{h}" for h in range(0,vals.shape[-1])]
                skip_first = 0
            else:
                n_layers, n_heads = vals.shape[0], vals.shape[2]
                vals = np.swapaxes(vals,1,2)
                vals = vals.reshape(model.cfg.n_layers*model.cfg.n_heads, -1)
                yticklabels = []
                skip_first = 1
                for l in range(0, n_layers):
                    for h in range(0, n_heads):
                        yticklabels.append(f"L{l}H{h}")
                        
    else:
        vals = SCALARS[c_type].detach().squeeze().numpy()
        yticklabels = [f"L{l}" for l in range(0, vals.shape[0])]
        xticklabels = model.to_str_tokens(toks_nI)
        skip_first = 1
    
    
    ##_____________________________________________
    
    fig, ax = plt.subplots(1, 1, figsize=figsize)    
    s = sns.heatmap(vals[:,skip_first:], #annot=rankdata(vals, axis=-1)-1
                  cmap=mpl.colormaps["coolwarm"], center=0.0,
                  xticklabels=xticklabels[skip_first:],
                  yticklabels=yticklabels, square=False, cbar=True,
                  cbar_kws={'location': 'top', 'shrink':0.8, 'pad': 0.01,'label':f''}, ax=ax) 
    #sns.set(font_scale=1.0)
    s.tick_params(left=True, bottom=True)
    
    if show_every==-1:
        ax.axes.get_yaxis().set_visible(False)
    else:
        ax.set_yticks(np.arange(skip_first, len(yticklabels), show_every))  # Set x-axis ticks to every other tick
        ax.set_yticklabels(yticklabels[skip_first::show_every])  # Set x-axis labels to every other tick
    plt.xticks(rotation=90)
    
    ax.invert_yaxis()
    title = f"{attrType}\non {c_type}"
    
    ax.set_xlabel(f"{model.to_str_tokens(answ_toks_n2.view(-1))}",fontsize=fontsize-1)
    #ax.set_ylabel(f"{title}",fontsize=fontsize)
    
    _ = ax.set_title(title, fontsize=fontsize, loc="left", y=1.23, x=0)
    fig.savefig(f"{dataHandler.ROOT}/results/{attrType}{model.to_str_tokens(answ_toks_n2)[0]}{f_name}.pdf", dpi=200, bbox_inches="tight")


def plot_beta_intervention(attrType, model, beta_range, beta_tok_logits, beta_tok_ranks, answ_toks_n2, f_name:str="", figsize=(4,3), fontsize=23):
    
    fig, ax = plt.subplots(1, 1, figsize=figsize)
    
    label0 = ax.plot(beta_range,beta_tok_logits[:,0], color="red", linewidth=2, label=model.to_str_tokens(answ_toks_n2)[0])
    label1 = ax.plot(beta_range,beta_tok_logits[:,1], color="blue", linewidth=2, label=model.to_str_tokens(answ_toks_n2)[1])
    
    for i, beta_x in enumerate(beta_range):
        plt.text(beta_x-0.01, beta_tok_logits[i,0]-0.2, str(beta_tok_ranks[i][0][0]), rotation=0, color="red", fontsize=fontsize)
        plt.text(beta_x-0.01, beta_tok_logits[i,1]+0.2, str(beta_tok_ranks[i][1][0]), rotation=0, color="blue", fontsize=fontsize)
    
    ax.text(0.0, 1.1, model.to_str_tokens(answ_toks_n2)[0], rotation=0, color="red", fontsize=fontsize, transform=ax.transAxes)
    ax.text(0.5, 1.1, model.to_str_tokens(answ_toks_n2)[1], rotation=0, color="blue", fontsize=fontsize, transform=ax.transAxes)
    
    ax.axvline(0, color="grey", linewidth=2, linestyle=":")
    ax.set_xlabel(f"beta",fontsize=fontsize, loc="center")
    ax.yaxis.set_major_formatter(FormatStrFormatter('%.f'))
    #ax.set_ylabel(f"token logit",fontsize=fontsize)
    ax.tick_params(axis='both', which='major', labelsize=fontsize)
    #ax.legend(handles =label0+label1, loc ='center left', fontsize=fontsize, frameon=False, bbox_to_anchor=(-0.25, 1.15), ncol=2) 
    
    #plt.text(-0.0, 1.3, model.to_string(toks_NI[1:]), fontsize=fontsize, rotation=0, wrap=True, ha='left', va='top', transform=ax.transAxes)
    #_ = ax.set_title(f"Token logit and rank", fontsize=fontsize, loc="left", y=1.2, x=-0.1)
    #_ = ax.set_title(f"{model.to_string(toks_NI[1:])}", fontsize=fontsize, loc="left") #{model.to_str_tokens(answ_toks_n2.view(-1))}
    
    fig.savefig(f"{dataHandler.ROOT}/results/{attrType}{model.to_str_tokens(answ_toks_n2)[0]} beta{f_name}.pdf", dpi=200, bbox_inches="tight")


